(******************************************************************************)
(* UCalcStatDesOrdinalAssociation.pas - Copyright (c) 2004 Ricco RAKOTOMALALA *)
(******************************************************************************)

{
@abstract(Structure pour calculer les dpendances entre variables ordinales)
@author(Ricco)
@created(28/09/2007)

On utilise pleinement le support de cours crit dans le mme tempo
--> http://eric.univ-lyon2.fr/~ricco/cours/cours/Dependance_Variables_Qualitatives.pdf 
}
unit UCalcStatDesOrdinalAssociation;

interface

USES
    UCalcStatDes,
    UCalcStructOrdinalAssociation,
    UDatasetExamples;

TYPE
    ///type de base
    TCalcSDOrdinalAssoc = class(TCalcStatDes)
    private
    //pas propritaire
    FY, FX: TVecDistinctValues;
    //tableau de calcul - propritaire
    FTab: TCrossOrdinalVariables;
    protected
    //indicateur et des ecarts-type
    FMeasure, FStdErr, FStdErrH0: double;
    //Z et la p-value
    FZ, FPValue: double;
    //bornes basses et haute  95%
    FLow, FUp: double;
    //lancer le calcul de l'indicateur
    procedure   calcIndicator(); virtual; abstract;
    //calcul de l'cart-type
    procedure    calcStdErr(); virtual; abstract;
    //cart-type sous H0
    procedure    calcStdErrH0(); virtual; abstract;
    //calcul du Z et de la p-value -- normalit asymptotique used
    procedure    calcZAndPValue();
    public
    //constructeur
    constructor create(prmY,prmX: TVecDistinctValues; prmExamples: TExamples = nil); virtual;
    //destructor
    destructor  destroy(); override;
    //calcul des indicateurs, notamment des paires concordantes et discordantes
    procedure   RefreshStat(prmExamples: TExamples); override;
    //obtenir la valeur de l'indicateur
    function    getMeasure(): double;
    //obtenir la valeur absolue de l'indicateur
    function    getAbsMeasure(): double;
    //description des rsultats
    function    getHTMLResult(prmOption: integer = -1): string; override;
    end;

    //classe de classe
    TClassCalcSDOrdinalAssoc = class of TCalcSDOrdinalAssoc;

    ///liste de stats ordinales
    TLstSDOrdinalAssoc = class(TLstCalcStatDes)
    protected
    function    getEachStatHTMLResults(prmOption: integer = -1): string; override;
    public
    {comparaison pour les stats conditionnelles}
    procedure SortStats(); override;
    {description HTML des rsultats}
    function    getHeaderHTML(): string; override;
    end;

    //gamma de Goodman et Kruskal
    TOrdinalAssocGamma = class(TCalcSDOrdinalAssoc)
    protected
    procedure   calcIndicator(); override;
    procedure   calcStdErr(); override;
    procedure   calcStdErrH0(); override;
    end;

    //tau-b de Kendall
    TOrdinalAssocTau_b = class(TCalcSDOrdinalAssoc)
    protected
    procedure   calcIndicator(); override;
    procedure   calcStdErr(); override;
    procedure   calcStdErrH0(); override;
    end;

    //tau-c de Kendall
    TOrdinalAssocTau_c = class(TCalcSDOrdinalAssoc)
    protected
    procedure   calcIndicator(); override;
    procedure   calcStdErr(); override;
    procedure   calcStdErrH0(); override;
    end;

    //d de Sommers
    TOrdinalAssocSommers = class(TCalcSDOrdinalAssoc)
    protected
    procedure   calcIndicator(); override;
    procedure   calcStdErr(); override;
    procedure   calcStdErrH0(); override;
    end;

implementation

USES
    Sysutils, Classes, Math,
    FMath,
    UConstConfiguration;

CONST
    //quantile d'ordre 0.975 pour la loi normale
    U975 : double = 1.9599627874;

{ TCalcSDOrdinalAssoc }

procedure TCalcSDOrdinalAssoc.calcZAndPValue;
begin
 if (FStdErrH0 > 0.0) then FZ:= FMeasure / FStdErrH0 else FZ := 0.0;
 FPValue:= PNorm(FZ); 
end;

constructor TCalcSDOrdinalAssoc.create(prmY, prmX: TVecDistinctValues;
  prmExamples: TExamples);
begin
 inherited create(nil,nil);
 FY:= prmY; FX:= prmX;
 FTab:= TCrossOrdinalVariables.create(FY,FX,nil); 
 if (prmExamples <> nil) then self.RefreshStat(prmExamples);
end;

destructor TCalcSDOrdinalAssoc.destroy;
begin
 if (FTab <> nil) then FreeAndNil(FTab);
 inherited;
end;

function TCalcSDOrdinalAssoc.getAbsMeasure: double;
begin
 result:= abs(self.getMeasure());
end;

function TCalcSDOrdinalAssoc.getHTMLResult(prmOption: integer): string;
var s: string;
begin
 s:= '';
 //si dtails demands, ajouter la description du tableau de contingence
 if (prmOption = 1)
  then
   begin
    s:= s + HTML_TABLE_COLOR_DATA_GRAY + '<TD align="left" colspan=8>';
    s:= s + FTab.getHTMLDescription();
    s:= s + '</TD></TR>';
   end;

 //dcrire les rsultats
 s:=  s + HTML_TABLE_COLOR_DATA_BLUE  + format('<TD align="right">%s</TD><TD align="right">%s</TD><TD align="right">%.6f</TD><TD align="right">%.6f</TD><TD align="right">[%.4f ; %.4f]</TD><TD align="right">%.6f</TD><TD align="right">%.6f</TD><TD align="right">%.6f</TD></TR>',
             [FY.Attribute.Name,FX.Attribute.Name,FMeasure,FStdErr,FLow,FUp,FStdErrH0,FZ,FPValue]);

  result:= s;
end;

function TCalcSDOrdinalAssoc.getMeasure: double;
begin
 result:= FMeasure;
end;

procedure TCalcSDOrdinalAssoc.RefreshStat(prmExamples: TExamples);
begin
 //mj des tables
 FTab.refreshTables(prmExamples);
 FNbExamples:= prmExamples.Size;
 //lancer le calcul de l'indicateur
 self.calcIndicator();
 //de ses ecarts-type
 self.calcStdErr();
 self.calcStdErrH0();
 //et de Z, p-value
 self.calcZAndPValue();
 //les bornes basses et hautes de l'intervalle de confiance
 FLow:= FMeasure - U975 * FStdErr;
 FUp:= FMeasure + U975 * FStdErr;
end;

{ TLstSDOrdinalAssoc }

function TLstSDOrdinalAssoc.getEachStatHTMLResults(
  prmOption: integer): string;
var i: integer;
    stat: TCalcStatDes;
    s: string;
begin
 s:= '';
 for i:= 0 to pred(self.Count) do
  begin
   //chaque ligne de rsultat du tableau
   stat:= self.Stat(i);
   s:= s+stat.getHTMLResult(prmOption);
  end;
 result:= s;
end;

function TLstSDOrdinalAssoc.getHeaderHTML: string;
begin
 result:=  HTML_TABLE_COLOR_HEADER_BLUE+
           '<TH width=80>Y</TH><TH width=80>X</TH><TH width=60>Measure</TH><TH width=60>Std-err</TH><TH width=80>95-percent C.I.</TH><TH width=70>Std-Err (H0)</TH><TH width=60>z</TH><TH width=60>Pr(>|z|)</TH>' + '</TR>';
end;

{trier selon le nom des attributs en ligne}
function ListSortCompareRowAttName(item1,item2: pointer): integer;
var st1,st2: TCalcSDOrdinalAssoc;
begin
 st1:= TCalcSDOrdinalAssoc(item1);
 st2:= TCalcSDOrdinalAssoc(item2);
 if (st1.FY.Attribute.Name < st2.FY.Attribute.Name)
  then result:= -1
  else
   if (st1.FY.Attribute.Name > st2.FY.Attribute.Name)
    then result:= +1
    else result:= 0;
end;

{trier selon le nom des attributs en colonne}
function ListSortCompareColumnAttName(item1,item2: pointer): integer;
var st1,st2: TCalcSDOrdinalAssoc;
begin
 st1:= TCalcSDOrdinalAssoc(item1);
 st2:= TCalcSDOrdinalAssoc(item2);
 if (st1.FX.Attribute.Name < st2.FX.Attribute.Name)
  then result:= -1
  else
   if (st1.FX.Attribute.Name > st2.FX.Attribute.Name)
    then result:= +1
    else result:= 0;
end;

{trier selon le nom des attributs en colonne}
function ListSortCompareMeasure(item1,item2: pointer): integer;
var st1,st2: TCalcSDOrdinalAssoc;
begin
 st1:= TCalcSDOrdinalAssoc(item1);
 st2:= TCalcSDOrdinalAssoc(item2);
 if (st1.getMeasure() > st2.getMeasure())
  then result:= -1
  else
   if (st1.getMeasure() < st2.getMeasure())
    then result:= +1
    else result:= 0;
end;

{trier selon le nom des attributs en colonne}
function ListSortCompareAbsMeasure(item1,item2: pointer): integer;
var st1,st2: TCalcSDOrdinalAssoc;
begin
 st1:= TCalcSDOrdinalAssoc(item1);
 st2:= TCalcSDOrdinalAssoc(item2);
 if (st1.getAbsMeasure() > st2.getAbsMeasure())
  then result:= -1
  else
   if (st1.getAbsMeasure() < st2.getAbsMeasure())
    then result:= +1
    else result:= 0;
end;

procedure TLstSDOrdinalAssoc.SortStats;
var funcCompare: TListSortCompare;
begin
 if (CompareMode>=0)
  then
   begin
    case CompareMode of
     0: funcCompare:= ListSortCompareRowAttName;
     1: funcCompare:= ListSortCompareColumnAttName;
     2: funcCompare:= ListSortCompareMeasure;
     3: funcCompare:= ListSortCompareAbsMeasure;
     else
      funcCompare:= NIL;
    end;

    if assigned(funcCompare) then LstStat.Sort(funcCompare);
   end;
end;

{ TOrdinalAssocGamma }

procedure TOrdinalAssocGamma.calcIndicator;
var s: double;
begin
 s:= FTab.P + FTab.Q;
 if (s > 0.0)
  then FMeasure:= (FTab.P - FTab.Q) / s
  else FMeasure:= 0.0;
end;

procedure TOrdinalAssocGamma.calcStdErr;
var v: double;
    l,c: integer;
begin
 v:= 0.0;
 for l:= 1 to FY.nbValues do
  for c:= 1 to FX.nbValues do
   v:= v + FTab.Values[l,c] * power(FTab.Q * FTab.Clc[l,c] - FTab.P * FTab.Dlc[l,c],2.0);
 if (FTab.P + FTab.Q > 0.0) then v:= 16.0/power(FTab.P + FTab.Q,4.0) * v else v:= 0.0;
 //racine pour l'cart-type
 FStdErr:= sqrt(v);
end;

procedure TOrdinalAssocGamma.calcStdErrH0;
var v: double;
    l,c: integer;
begin
 v:= 0.0;
 for l:= 1 to FY.nbValues do
  for c:= 1 to FX.nbValues do
   v:= v + FTab.Values[l,c] * power(FTab.Clc[l,c] - FTab.Dlc[l,c],2.0);
 v:= v - 1.0 / FTab.n * power(Ftab.P - FTab.Q,2.0);
 if (FTab.P + FTab.Q > 0.0) then v:= 4.0/power(FTab.P + FTab.Q,2.0) * v else v:= 0.0;
 //racine pour l'cart-type
 FStdErrH0:= sqrt(v);
end;

{ TOrdinalAssocTau_b }

procedure TOrdinalAssocTau_b.calcIndicator;
begin
 if (FTab.DY * FTab.DX > 0.0)
  then FMeasure:= (FTab.P - FTab.Q)/sqrt(FTab.DY * FTab.DX)
  else FMeasure:= 0.0;
end;

procedure TOrdinalAssocTau_b.calcStdErr;
var tmp, tmp1, tmp2: double;
    v: double;
    l,c: integer;
begin
 if (FTab.DY * FTab.DX > 0.0)
  then
   begin
    //produit sqrt(DY * DX)
    tmp1 := sqrt(FTab.DY * FTab.DX);
    //construire la double somme
    v:= 0.0;
    for l:= 1 to FY.nbValues do
     for c:= 1 to FX.nbValues do
      begin
       //nu_{lc}
       tmp2:= FTab.Values[l,0] * FTab.DX + FTab.Values[0,c] * FTab.DY;
       //valeur dans la double somme
       tmp:= power(2.0 * tmp1 * (FTab.Clc[l,c] - FTab.Dlc[l,c]) + FMeasure * tmp2, 2.0);
       //sommer
       v:= v + FTab.Values[l,c] * tmp;
      end;
    //retrancher et diviser
    v:= v - power(FTab.n,3.0) * power(FMeasure,2.0) * power(FTab.DY + FTab.DX, 2.0);
    v:= v / power(FTab.DY * FTab.DX, 2.0);
    //ecart-type
    if (v > 0.0) then FStdErr:= sqrt(v) else FStdErr:= 0.0;
   end
  else FStdErr:= 0.0;
end;

procedure TOrdinalAssocTau_b.calcStdErrH0;
var v: double;
    l,c: integer;
begin
 if (FTab.DY * FTab.DX > 0.0)
  then
   begin
    //construire la double somme
    v:= 0.0;
    for l:= 1 to FY.nbValues do
     for c:= 1 to FX.nbValues do
      v:= v + FTab.Values[l,c] * power(FTab.Clc[l,c] - FTab.Dlc[l,c], 2.0); 
    //retrancher et diviser
    v:= v - 1.0 / FTab.n * power(FTab.P - FTab.Q,2.0);
    v:= 4.0 * v / (FTab.DY * FTab.DX);
    //ecart-type
    if (v > 0.0) then FStdErrH0:= sqrt(v) else FStdErrH0:= 0.0;
   end
  else FStdErrH0:= 0.0;
end;

{ TOrdinalAssocTau_c }

procedure TOrdinalAssocTau_c.calcIndicator;
var q: double;
begin
 q:= 1.0 * math.Min(FY.nbValues,FX.nbValues);
 if (q > 1.0)
  then FMeasure:= q * (FTab.P - FTab.Q) / (power(FTab.n,2.0) * (q - 1.0))
  else FMeasure:= 0.0;
end;

procedure TOrdinalAssocTau_c.calcStdErr;
var v,q: double;
    l,c: integer;
begin
 q:= 1.0 * math.Min(FY.nbValues,FX.nbValues);
 if (q > 1.0)
  then
   begin
    //construire la double somme
    v:= 0.0;
    for l:= 1 to FY.nbValues do
     for c:= 1 to FX.nbValues do
      v:= v + FTab.Values[l,c] * power(FTab.Clc[l,c] - FTab.Dlc[l,c], 2.0); 
    //retrancher et diviser
    v:= v - 1.0 / FTab.n * power(FTab.P - FTab.Q,2.0);
    v:= 4.0 * power(q,2.0) / (power(FTab.n,4.0) * power(q-1.0,2.0)) * v;
    //ecart-type
    if (v > 0.0) then FStdErr:= sqrt(v) else FStdErr:= 0.0;
   end
  else FStdErr:= 0.0;
end;

procedure TOrdinalAssocTau_c.calcStdErrH0;
begin
 //h ben dis donc, pour une fois...
 FStdErrH0:= FStdErr;
end;

{ TOrdinalAssocSommers }

procedure TOrdinalAssocSommers.calcIndicator;
begin
 if (FTab.DX > 0.0)
  then FMeasure:= (FTab.P - FTab.Q)/FTab.DX
  else FMeasure:= 0.0;
end;

procedure TOrdinalAssocSommers.calcStdErr;
var v: double;
    l,c: integer;
begin
 if (FTab.Dx > 0.0)
  then
   begin
    //construire la double somme
    v:= 0.0;
    for l:= 1 to FY.nbValues do
     for c:= 1 to FX.nbValues do
      v:= v + FTab.Values[l,c] * power(FTab.DX * (FTab.Clc[l,c] - FTab.Dlc[l,c]) - (FTab.P - FTab.Q) * (FTab.n - FTab.Values[0,c]), 2.0); 
    //retrancher et diviser
     v:= 4.0 / power(FTab.DX,4.0) * v;
    //ecart-type
    if (v > 0.0) then FStdErr:= sqrt(v) else FStdErr:= 0.0;
   end
  else FStdErr:= 0.0;
end;

procedure TOrdinalAssocSommers.calcStdErrH0;
var v: double;
    l,c: integer;
begin
 if (FTab.DX > 0.0)
  then
   begin
    //construire la double somme
    v:= 0.0;
    for l:= 1 to FY.nbValues do
     for c:= 1 to FX.nbValues do
      v:= v + FTab.Values[l,c] * power(FTab.Clc[l,c] - FTab.Dlc[l,c], 2.0); 
    //retrancher et diviser
    v:= v - 1.0 / FTab.n * power(FTab.P - FTab.Q,2.0);
    v:= 4.0 * v / power(FTab.DX,2.0);
    //ecart-type
    if (v > 0.0) then FStdErrH0:= sqrt(v) else FStdErrH0:= 0.0;
   end
  else FStdErrH0:= 0.0;
end;

end.
