(**************************************************************)
(* UCalcFSReliefF.pas - Copyright (c) 2010 Ricco RAKOTOMALALA *)
(**************************************************************)

{
@abstract(Mthode de slection de variables RELIEFF)
@author(Ricco)
@created(17/03/2010)

Implmente la liste des attributs pour la mthode de slection de variables suprvise RELIEFF (multi-classes).
Rf. "Overcoming the myopia of inductive learning algorithms with RELIEFF", Kononenko et al., 1997.

Paramtres :

- K nombre de voisins (dans chaque classe)
- m taille de l'chantillon extrait (tirage avec remise)
- seed valeur de dpart du gnrateur de nombre alatoire
- mode de slection des variables
 >> nombre prdfini
 >> poids > 0
 >> poids > moyenne des poids positifs (dfaut)

}

unit UCalcFSReliefF;

interface

USES
    UDatasetDefinition, UDatasetExamples,
    Contnrs, UCompFSRelieff;

TYPE
    //**
    //attribu de type RELIEFF
    TAttRelieff = class
    public
    //attribut associ
    Att: TAttribute;
    //poids associ
    Weight: double;
    //constructeur
    constructor create(prmAtt: TAttribute; examples: TExamples);
    //distance normalise entre deux points
    function    normDist(example, exRef: integer): double; virtual; abstract;
    end;

    //**
    //attribut de type discret
    TAttRelieffDiscrete = class(TAttRelieff)
    public
    function    normDist(example, exRef: integer): double; override;
    end;

    //**
    //attribut de type discret
    TAttRelieffContinuous = class(TAttRelieff)
    private
    FMin, FMax, FRange: double;
    public
    constructor create(prmAtt: TAttribute; examples: TExamples);
    function    normDist(example, exRef: integer): double; override;
    end;

    //***
    //Liste des attributs relieff
    TLstAttRelieff = class
    public
    //liste interne
    Lst: TObjectList;
    //constructeur
    constructor create();
    //destructeur
    destructor  destroy(); override;
    //ajouter un attribut
    procedure   add(att: TAttRelieff);
    //obtenir un attribut relief
    function    getAtt(index: integer): TAttRelieff;
    //nombre d'lments
    function    count(): integer;
    //trier selon un poids dcroissant
    procedure   sort();
    //vider -- attention, la liste est propritaire
    procedure   clear();
    //cart entre deux observations en prenant en compte toutes les variables
    function   distance(example, exRef: integer): double;
    end;

    //dclaration forward
    TCalcRelieff = class;

    //**
    //index pour une classe
    TIndexClassValue = class
    public
    IdClasse: TTypeDiscrete;
    NbExamples: integer;
    Index: TExamples;//index construit ininitalement, jamais remis en cause
    SortedIndex: TExamples;//index tri - recalcul pour chaque nouvel individu
    AttIndex: TAttribute;//vecteur des distances, astuce pour pouvoir trier constamment
    //
    constructor create(individus: TExamples; classe: TTypeDiscrete);
    destructor  destroy(); override;//tout est  dtruire
    //distance d'un exemple avec TOUS les individus de la liste + mise  jour des poids
    procedure   distances(example: integer; calc: TCalcRelieff);
    end;

    //**
    //liste des index par classe
    TLstIndex = class
    private
    //maj hit
    function     majHit(idHit: TTypeDiscrete; att: TAttRelieff; calc: TCalcRelieff; example: integer): double;
    //mj miss
    function     majMisses(idHit: TTypeDiscrete; att: TAttRelieff; calc: TCalcRelieff; example: integer): double;
    //obtenir les poids des K voisins pour un attribut
    function     getPoids(idClasse: TTypeDiscrete; att: TAttRelieff; calc: TCalcRelieff; example: integer): double;
    public
    LstIndex: TObjectList;//liste interne des index
    constructor create();
    destructor  destroy(); override;
    procedure   addIndex(index: TIndexClassValue);
    function    getIndex(idClasse: TTypeDiscrete): TIndexClassValue;
    procedure   process(example: integer; calc: TCalcRelieff);
    end;

    //**
    //classe de calcul RELIEFF
    TCalcRelieff = class
    private
    //paramtres de calcul
    FPrm: TOpPrmFSRelieff;
    //calculer les poids
    function     computeWeight(examples: TExamples): boolean;
    public
    //Nombre d'observations
    NbExamples: integer;
    //nombre d'observations utilise pour recalculer les poids
    UsedNbExamples: integer;
    //variable cible
    Target: TAttribute;
    //liste des attributs de type Relieff
    LstAtt: TLstAttRelieff;
    //constructeur
    constructor create(prm: TOpPrmFSRelieff);
    //destructeur
    destructor  destroy(); override;
    //apprentissage
    function    learning(operator: TOpFSRelieff; examples: TExamples): boolean;
    end;



implementation

USES
    Math, UDatasetImplementation, UCalcRndGenerator, SysUtils, ULogFile;

{ TAttRelieff }

constructor TAttRelieff.create(prmAtt: TAttribute; examples: TExamples);
begin
 inherited create();
 att:= prmAtt;
 Weight:= 0.0;
end;

{ TAttRelieffDiscrete }

function TAttRelieffDiscrete.normDist(example, exRef: integer): double;
begin
 result:= ord(att.dValue[example] <> att.dValue[exRef]);
end;

{ TAttRelieffContinuous }

constructor TAttRelieffContinuous.create(prmAtt: TAttribute;
  examples: TExamples);
var i: integer;
    value: double;
begin
 inherited create(prmAtt,examples);
 FMax:= -1.0 * Math.MaxDouble;
 FMin:= Math.MaxDouble;
 for i:= 1 to examples.Size do
  begin
   value:= att.cValue[examples.Number[i]];
   if (value > FMax) then FMax:= value;
   if (value < FMin) then FMin:= value;
  end;
 FRange:= FMax - FMin;
end;

function TAttRelieffContinuous.normDist(example, exRef: integer): double;
var value: double;
begin
 if (FRange > 0)
  then
   begin
    //cart absolu
    value:= ABS(att.cValue[example] - att.cValue[exRef])/FRange;
    //si dbordement les min max obtenus sur l'chantillon d'apprentissage
    value:= Math.Min(value,1.0);
    value:= Math.Max(value,0.0);
    //
    result:= value;
   end
  else result:= 0.0;
end;

{ TLstAttRelieff }

procedure TLstAttRelieff.add(att: TAttRelieff);
begin
 Lst.Add(att);
end;

procedure TLstAttRelieff.clear;
begin
 Lst.Clear();
end;

function TLstAttRelieff.count: integer;
begin
 result:= Lst.Count;
end;

constructor TLstAttRelieff.create;
begin
 inherited create();
 Lst:= TObjectList.Create(TRUE);
end;

destructor TLstAttRelieff.destroy;
begin
 Lst.Free();
 inherited;
end;

function TLstAttRelieff.distance(example, exRef: integer): double;
var j: integer;
    value: double;
    att: TAttRelieff;
begin
 value:= 0.0;
 for j:= 0 to pred(Lst.Count) do
  begin
   att:= self.getAtt(j);
   value:= value + att.normDist(example,exRef);
  end;
 result:= value;
end;

function TLstAttRelieff.getAtt(index: integer): TAttRelieff;
begin
 result:= Lst.Items[index] as TAttRelieff;
end;

//comparaison selon le poids
function compareWeight(item1, item2: Pointer): integer;
var obj1, obj2: TAttRelieff;
begin
 obj1:= TAttRelieff(item1);
 obj2:= TAttRelieff(item2);
 //tri dcroissant
 if (obj1.Weight < obj2.Weight)
  then result:= +1
  else
   if (obj1.Weight > obj2.Weight)
    then result:= -1
    else result:= 0;
end;
//

procedure TLstAttRelieff.sort;
begin
 Lst.Sort(compareWeight);
end;

{ TCalcRelieff }

function TCalcRelieff.computeWeight(examples: TExamples): boolean;
var rnd: TRndGenerator;
    i,k: integer;
    example: integer;
    tmpEx: TObjectList;
    KMin: integer;
    curEx: TExamples;
    LstIndex: TLstIndex;
begin
 RESULT:= TRUE;
 LstIndex:= TLstIndex.create();
 rnd:= TRndGenerator.Create(seedUser,FPrm.Seed,DEFAULT_SEED_VALUE_2);
 TRY
 //calculer les effectifs par classe
 tmpEx:= examples.DispatchExamples(self.Target);
 tmpEx.OwnsObjects:= FALSE;//modifier le statut de proprit
 //mj du nombre de voisins max. - on en profite pour crer la liste des index par classe
 KMin:= FPrm.K;
 for k:= 1 to self.Target.nbValues do
  begin
   curEx:= tmpEx.Items[pred(k)] as TExamples;
   if (curEx.Size < KMin) then KMin:= curEx.Size;
   //cration des listes par classe
   LstIndex.addIndex(TIndexClassValue.create(curEx,k));
  end;
 tmpEx.Free();//vider la liste temporaire, qui n'est plus propritaire
 //mj ventuelle du nombre de voisins  prendre en compte
 FPrm.K:= Min(FPrm.K,Kmin);
 //selon le mode d'chantillonnage
 If (FPrm.UseSampling = TRUE) and (FPrm.m > 0) then
  begin
   self.UsedNbExamples:= Fprm.m;
   //processus de mj pour chaque individu tir alatoirement
   for i:= 1 to FPrm.m do
    begin
     example:= examples.Number[succ(rnd.IRanMarRange(examples.Size))];
     LstIndex.process(example,self);
    end;
  end
 else
  begin
   self.UsedNbExamples:= examples.Size;
   for i:= 1 to examples.Size do LstIndex.process(examples.number[i],self);
  end;
 //puis trier finalement
 self.LstAtt.sort();
 EXCEPT
 RESULT:= FALSE;
 END;
 LstIndex.Free();
 rnd.Free();
end;

constructor TCalcRelieff.create(prm: TOpPrmFSRelieff);
begin
 inherited create();
 FPrm:= prm;
 LstAtt:= TLstAttRelieff.create();
end;

destructor TCalcRelieff.destroy;
begin
 LstAtt.Free();
 inherited;
end;

function TCalcRelieff.learning(operator: TOpFSRelieff; examples: TExamples): boolean;
var j: integer;
    att: TAttribute;
begin
 NbExamples:= examples.Size;
 //charger les variables
 self.Target:= operator.WorkData.LstAtts[asTarget].Attribute[0];
 //vider la liste
 LstAtt.clear();
 TRY
 //puis la charger
 for j:= 0 to pred(operator.WorkData.LstAtts[asInput].Count) do
  begin
   att:= operator.WorkData.LstAtts[asInput].Attribute[j];
   if (att.isCategory(caDiscrete) = true)
    then LstAtt.add(TAttRelieffDiscrete.create(att,examples))
    else LstAtt.add(TAttRelieffContinuous.create(att,examples));
  end;
 //construire les poids maintenant puis trier selon l'ordre d'importance...
 RESULT:= self.computeWeight(examples);
 EXCEPT
 RESULT:= FALSE;
 END;
end;

{ TIndexClassValue }

constructor TIndexClassValue.create(individus: TExamples;
  classe: TTypeDiscrete);
begin
 inherited create();
 Index:= individus;
 NbExamples:= individus.Size;
 SortedIndex:= TExamples.Create(individus.Size);
 idClasse:= classe;
 AttIndex:= TAttContinue.Create(IntToStr(idClasse),individus.Size);
end;

destructor TIndexClassValue.destroy;
begin
 Index.Free();
 SortedIndex.Free();
 AttIndex.Free();
 inherited;
end;

procedure TIndexClassValue.distances(example: integer;
  calc: TCalcRelieff);
var i,exRef: integer;
    dist: double;
begin
 //calculer les distances
 for i:= 1 to Index.Size do
  begin
   exRef:= Index.Number[i];
   dist:= calc.LstAtt.distance(example,exRef);
   AttIndex.cValue[i]:= dist;
  end;
 //trier les index selon les valeurs
 SortedIndex.fillValues();
 SortedIndex.QuickSortBy(AttIndex);
end;

{ TLstIndex }

procedure TLstIndex.addIndex(index: TIndexClassValue);
begin
 LstIndex.Add(index);
end;

constructor TLstIndex.create;
begin
 inherited create();
 LstIndex:= TObjectList.Create(TRUE);
end;

destructor TLstIndex.destroy;
begin
 LstIndex.Free();
 inherited;
end;

function TLstIndex.getIndex(idClasse: TTypeDiscrete): TIndexClassValue;
begin
 result:= LstIndex.Items[pred(idClasse)] as TIndexClassValue;
end;

procedure TLstIndex.process(example: integer; calc: TCalcRelieff);
var k,j: integer;
    idClasse: TTypeDiscrete;
    att: TAttRelieff;
    hit,miss: double;
begin
 //pour chaque classe, calculer la distance  l'individu
 for k:= 1 to calc.Target.nbValues do self.getIndex(k).distances(example,calc);
 //maintenant, il s'agit de mettre  jour les poids
 idClasse:= calc.Target.dValue[example];
 //pour chaque attribut
 for j:= 0 to pred(calc.LstAtt.count) do
  begin
   att:= calc.LstAtt.getAtt(j);
   //m des poids selon les HIT et les MISSES
   hit:= self.majHit(idClasse,att,calc,example);
   miss:= self.majMisses(idClasse,att,calc,example);
   att.Weight:= att.Weight - hit + miss;
   //TraceLog.WriteToLogFile(format('[%s] --> hit : %.4f, miss : %.4f, weigth = %.4f',[att.Att.Name,hit,miss,att.Weight]));
  end;
end;

function TLstIndex.majHit(idHit: TTypeDiscrete; att: TAttRelieff;
  calc: TCalcRelieff; example: integer): double;
var poids: double;
begin
 poids:= self.getPoids(idHit,att,calc,example);
 result:= poids;
end;

function TLstIndex.majMisses(idHit: TTypeDiscrete; att: TAttRelieff;
  calc: TCalcRelieff; example: integer): double;
var k: integer;
    pC,pR: double;
    poids,SumPoids: double;
begin
 //poids de la classe  analyser
 pR:= (1.0 * self.getIndex(idHit).NbExamples)/(1.0 * calc.NbExamples);
 //pour chaque classe
 SumPoids:= 0.0;
 for k:= 1 to calc.Target.nbValues do
  begin
   if (k <> idHit)
    then
     begin
      poids:= self.getPoids(k,att,calc,example);
      pC:= (1.0 * self.getIndex(k).NbExamples)/(1.0 * calc.NbExamples);
      SumPoids:= SumPoids + pC/(1.0-pR)*poids;
     end;
  end;
 //et au final
 result:= SumPoids;
end;

function TLstIndex.getPoids(idClasse: TTypeDiscrete; att: TAttRelieff;
  calc: TCalcRelieff; example: integer): double;
var poids: double;
    i: integer;
    index: TIndexClassValue;
    diff: double;
    exRef: integer;
    nbK: integer;
begin
 //indices
 index:= self.getIndex(idClasse);
 //poids cumul
 poids:= 0.0;
 //pour les K les plus proches
 i:= 0;
 nbK:= 0;
 while (nbK < calc.FPrm.K) and (i < index.NbExamples) do
  begin
   inc(i);
   exRef:= index.Index.Number[index.SortedIndex.Number[i]];
   if (exRef <> example)
    then
     begin
      diff:= att.normDist(example,exRef);
      poids:= poids + diff;
      inc(nbK);
     end;
  end;
 //poids normalis
 poids:= poids / (1.0 * calc.UsedNbExamples * nbK);
 //
 result:= poids;
end;

end.
