(**************************************************************)
(* UCompFSRelieff.pas - Copyright (c) 2010 Ricco RAKOTOMALALA *)
(**************************************************************)

{
@abstract(Composant de slection de variables - Base sur la mthode RELIEFF)
@author(Ricco)
@created(18/03/2010)

Implmente la liste des attributs pour la mthode de slection de variables suprvise RELIEFF (multi-classes).
Rf. "Overcoming the myopia of inductive learning algorithms with RELIEFF", Kononenko et al., 1997.

Paramtres :

- K nombre de voisins (dans chaque classe)
- m taille de l'chantillon extrait (tirage avec remise)
- seed valeur de dpart du gnrateur de nombre alatoire
- mode de slection des variables
 >> nombre prdfini
 >> poids > seuil
 >> poids > moyenne des poids positifs (dfaut)

}

unit UCompFSRelieff;

interface

USES
        Forms, Classes, IniFiles,
        UCompDefinition,
        UCompFSDefinition,
        UCompFSInputSelection,
        UOperatorDefinition;

TYPE
        {gnrateur}
        TMLGenFSRelieff = class(TMLGenFS)
        public
        function    GetClassMLComponent: TClassMLComponent; override;
        end;

        {composant}
        TMLCompFSRelieff = class(TMLCompFSInputSelection)
        protected
        function    getClassOperator: TClassOperator; override;
        end;

        {oprateur}
        TOpFSRelieff = class(TOperatorFSInputSelSupervised)
        private
        procedure   filterOnBest();
        procedure   filterOnThreshold();
        procedure   filterOnMean();
        protected
        //oprateur de calcul
        FCalc: TObject;
        function    getClassParameter: TClassOperatorParameter; override;
        public
        constructor Create(AOwner: TObject); override;
        destructor  Destroy(); override;
        function    getHTMLResultsSummary(): string; override;
        function    CoreExecute(): boolean; override;
        end;

        TModeSelVarRelieff = (relieffNbBest, relieffThreshold, relieffMean);

        {paramtrage}
        TOpPrmFSRelieff = class(TOpPrmFSInputSelection)
        protected
        procedure   SetDefaultParameters(); override;
        function    CreateDlgParameters(): TForm; override;
        public
        //nombre de voisins - 10
        K: integer;
        //Mode de constitution de l'chantillon - dfaut FALSE
        UseSampling: boolean; 
        //taille de l'chantillon - 100
        m: integer;
        //seed (valeur de dpart) du gnrateur de nombre alatoire - 1967
        Seed: integer;
        //mode de slection des variables  partir du ranking
        ModeSel: TModeSelVarRelieff;
        //diffrents seuils
        NbBest: integer; //5 par dfaut
        Threshold: double; //0.0 par dfaut
        //suite - I/O
        procedure   LoadFromStream(prmStream: TStream); override;
        procedure   SaveToStream(prmStream: TStream); override;
        procedure   LoadFromINI(prmSection: string; prmINI: TMemIniFile); override;
        procedure   SaveToINI(prmSection: string; prmINI: TMemIniFile); override;
        //description des paramtres
        function    getHTMLParameters(): string; override;
        end;


implementation

USES
    Sysutils, Math,
    UCalcFSRelieff, UConstConfiguration, UStringAddBuffered,
    UDatasetImplementation, UDlgOpPrmFSRelieff;

{ TMLGenFSRelieff }

function TMLGenFSRelieff.GetClassMLComponent: TClassMLComponent;
begin
 result:= TMLCompFSRelieff;
end;

{ TMLCompFSRelieff }

function TMLCompFSRelieff.getClassOperator: TClassOperator;
begin
 result:= TOpFSRelieff;
end;

{ TOpFSRelieff }

function TOpFSRelieff.CoreExecute: boolean;
var calc: TCalcRelieff;
begin
 if (FCalc <> nil) then FreeAndNil(FCalc);
 calc:= TCalcRelieff.create(self.PrmOp as TOpPrmFSRelieff);
 TRY
  RESULT:= calc.learning(self,self.WorkData.Examples);
  if (result <> true)
   then calc.Free()
   else
    begin
     FCalc:= calc;
     //remplir les sorties du composant
     //target
     self.OutputData.LstAtts[asTarget].Clear();
     self.OutputData.LstAtts[asTarget].Add(self.WorkData.LstAtts[asTarget].Attribute[0]);
     //les input
     self.OutputData.LstAtts[asInput].Clear();
     case (self.PrmOp as TOpPrmFSRelieff).ModeSel of
      relieffNbBest: self.filterOnBest();
      relieffThreshold: self.filterOnThreshold();
      else self.filterOnMean();
     end;
    end;
 EXCEPT
  RESULT:= FALSE;
 END;

end;

constructor TOpFSRelieff.Create(AOwner: TObject);
begin
 inherited;
end;

destructor TOpFSRelieff.Destroy;
begin
 if assigned(FCalc) then FCalc.Free();
 inherited;
end;

procedure TOpFSRelieff.filterOnBest;
var prm: TOpPrmFSRelieff;
    calc: TCalcRelieff;
    j: integer;
begin
 prm:= self.PrmOp as TOpPrmFSRelieff;
 calc:= FCalc as TCalcRelieff;
 if (prm.NbBest > 0)
  then
   begin
    for j:= 0 to pred(Math.Min(prm.NbBest,self.WorkData.LstAtts[asInput].Count)) do
     self.OutputData.LstAtts[asInput].Add(calc.LstAtt.getAtt(j).Att);
   end;
end;

procedure TOpFSRelieff.filterOnMean;
var calc: TCalcRelieff;
    attR: TAttRelieff;
    j: integer;
    mean,nbMean: double;
begin
 calc:= FCalc as TCalcRelieff;
 //calculer la moyenne
 j:= 0; mean:= 0.0; nbMean:= 0.0;
 repeat
  attR:= calc.LstAtt.getAtt(j);
  if (attR.Weight > 0)
   then
    begin
     mean:= mean + attR.Weight;
     nbMean:= nbMean + 1.0;
    end;
   inc(j);
 until (attR.Weight <= 0) or (j >= calc.LstAtt.count);
 if (nbMean > 0) then mean:= mean / nbMean else mean:= 0.0;
 //slectionner maintenant en se basant sur la moyenne
 j:= 0;
 REPEAT
  attR:= calc.LstAtt.getAtt(j);
  if (attR.Weight >= mean) then self.OutputData.LstAtts[asInput].Add(attR.Att);
  inc(j);
 UNTIL (attR.Weight < mean) or (j >= calc.LstAtt.Count);
end;

procedure TOpFSRelieff.filterOnThreshold;
var prm: TOpPrmFSRelieff;
    calc: TCalcRelieff;
    attR: TAttRelieff;
    j: integer;
begin
 prm:= self.PrmOp as TOpPrmFSRelieff;
 calc:= FCalc as TCalcRelieff;
 j:= 0;
 REPEAT
  attR:= calc.LstAtt.getAtt(j);
  if (attR.Weight >= prm.Threshold) then self.OutputData.LstAtts[asInput].Add(attR.Att);
  inc(j);
 UNTIL (attR.Weight < prm.Threshold) or (j >= calc.LstAtt.Count);
end;

function TOpFSRelieff.getClassParameter: TClassOperatorParameter;
begin
 result:= TOpPrmFSRelieff;
end;

function TOpFSRelieff.getHTMLResultsSummary: string;
var buf: TBufString;
    s: string;
    calc: TCalcRelieff;
    att: TAttRelieff;
    j: integer;
begin
 //branchement
 calc:= FCalc as TCalcRelieff;
 //
 buf:= TBufString.Create();
 buf.BeginUpdate();
 //
 //le nombre des attributs filtrs
 s:= '<H3>INPUT attribute selection</H3>';
 s:= s+HTML_HEADER_TABLE_RESULT;
 s:= s+HTML_TABLE_COLOR_HEADER_BLUE+'<TH colspan="2">INPUT selection</TH>';
 s:= s+format('%s<TD>Before filtering</TD><TD align="right" width="50">%d</TD></TR>',[HTML_TABLE_COLOR_DATA_BLUE,self.workdata.LstAtts[asInput].Count]);
 s:= s+format('%s<TD>After filtering</TD><TD align="right" width="50">%d</TD></TR>',[HTML_TABLE_COLOR_DATA_BLUE,self.outputdata.LstAtts[asInput].Count]);
 s:= s+'</table>';
 buf.AddStr(s);
 //ajouter les lments dans un tableau
 s:= '<H3>Ranked Attributes (Weight)</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY;
 s:= s + '<TH>N</TH><TH>Attribute</TH><TH>Weight</TH></TR>';
 buf.AddStr(s);
 for j:= 0 to pred(calc.LstAtt.count) do
  begin
   att:= calc.LstAtt.getAtt(j);
   s:= HTML_TABLE_COLOR_DATA_GRAY;
   if (j < self.OutputData.LstAtts[asInput].Count)
    then s:= s + format('<TD %s>%d</TD>',[HTML_BGCOLOR_DATA_GREEN,succ(j)])
    else s:= s + format('<TD>%d</TD>',[succ(j)]);
   s:= s + format('<TD width=150>%s</TD><TD align=right>%.6f</TD></TR>',[att.Att.Name,att.Weight]);
   buf.AddStr(s);
  end;
 buf.AddStr('</table>');
 //fin de l'ajout et exportation
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

{ TOpPrmFSRelieff }

function TOpPrmFSRelieff.CreateDlgParameters: TForm;
begin
 result:= TdlgOpPrmFSRelieff.CreateFromOpPrm(self);
end;

function TOpPrmFSRelieff.getHTMLParameters: string;
var s: string;
begin
 s:= HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY;
 s:= s + '<TH>Parameter</TH><TH>Value</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>%s</TD><TD align=right>%d</TD></TR>',['Neighbors',K]);
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>%s</TD><TD align=right>%d</TD></TR>',['Use sampling',ord(UseSampling)]);
 if (UseSampling = true)
  then
   begin
    s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>%s</TD><TD align=right>%d</TD></TR>',['Sample size',m]);
    s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>%s</TD><TD align=right>%d</TD></TR>',['Random seed',seed]);
   end;
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>%s</TD><TD align=right>%d</TD></TR>',['Selection mode',ord(ModeSel)]);
 if (ModeSel = relieffNbBest) then s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>%s</TD><TD align=right>%d</TD></TR>',['# selected',NbBest]);
 if (ModeSel = relieffThreshold) then s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD>%s</TD><TD align=right>%.2f</TD></TR>',['Threshold value',Threshold]);
 s:= s + '</table>';
 result:= s;
end;

procedure TOpPrmFSRelieff.LoadFromINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 K:= prmINI.ReadInteger(prmSection,'neighbors',K);
 UseSampling:= prmINI.ReadBool(prmSection,'use_sample',UseSampling);
 m:= prmINI.ReadInteger(prmSection,'sample_size',m);
 seed:= prmINI.ReadInteger(prmSection,'seed',Seed);
 ModeSel:= TModeSelVarRelieff(prmINI.ReadInteger(prmSection,'selection_mode',ord(ModeSel)));
 NbBest:= prmINI.ReadInteger(prmSection,'nb_best',NbBest);
 Threshold:= prmINI.ReadFloat(prmSection,'threshold',Threshold);
end;

procedure TOpPrmFSRelieff.LoadFromStream(prmStream: TStream);
begin
 prmStream.ReadBuffer(K,sizeof(K));
 prmStream.ReadBuffer(UseSampling,sizeof(UseSampling));
 prmStream.ReadBuffer(m,sizeof(m));
 prmStream.ReadBuffer(Seed,sizeof(Seed));
 prmStream.ReadBuffer(ModeSel,sizeof(ModeSel));
 prmStream.ReadBuffer(NbBest,sizeof(NbBest));
 prmStream.ReadBuffer(Threshold,sizeof(Threshold));
end;

procedure TOpPrmFSRelieff.SaveToINI(prmSection: string;
  prmINI: TMemIniFile);
begin
 prmINI.WriteInteger(prmSection,'neighbors',K);
 prmINI.WriteBool(prmSection,'use_sample',UseSampling);
 prmINI.WriteInteger(prmSection,'sample_size',m);
 prmINI.WriteInteger(prmSection,'seed',Seed);
 prmINI.WriteInteger(prmSection,'selection_mode',ord(ModeSel));
 prmINI.WriteInteger(prmSection,'nb_best',NbBest);
 prmINI.WriteFloat(prmSection,'threshold',Threshold);
end;

procedure TOpPrmFSRelieff.SaveToStream(prmStream: TStream);
begin
 prmStream.WriteBuffer(K,sizeof(K));
 prmStream.WriteBuffer(UseSampling,sizeof(UseSampling));
 prmStream.WriteBuffer(m,sizeof(m));
 prmStream.WriteBuffer(Seed,sizeof(Seed));
 prmStream.WriteBuffer(ModeSel,sizeof(ModeSel));
 prmStream.WriteBuffer(NbBest,sizeof(NbBest));
 prmStream.WriteBuffer(Threshold,sizeof(Threshold));
end;

procedure TOpPrmFSRelieff.SetDefaultParameters;
begin
 //nombre de voisins - 10
 K:= 10;
 //UseSampling
 UseSampling:= FALSE;
 //taille de l'chantillon - 200 (actif si UseSampling = TRUE)
 m:= 200;
 //seed (valeur de dpart) du gnrateur de nombre alatoire - 1967
 Seed:= 1967;
 //mode de slection des variables  partir du ranking
 ModeSel:= relieffMean;
 //diffrents seuils
 NbBest:= 5; //5 par dfaut
 Threshold:= 0.0; //0.0 par dfaut
end;

initialization
 RegisterClass(TMLGenFSRelieff);
end.
