(*********************************************************************)
(* UCalcSpvRuleInduction.pas - Copyright (c) 2009 Ricco RAKOTOMALALA *)
(*********************************************************************)

{
@abstract(Classe de calcul - Induction de rgles)
@author(Ricco)
@created(11/11/2009)

Inspir de CN2 de Clark et Boswell, 1991
Classe de calcul
}
unit UCalcSpvRuleInduction;

interface

USES
   UCompSpvLDefinition,
   UCalcDistributionEvaluation, UCalcRulesDefinition,
   UDatasetExamples, UDatasetDefinition, UCalcDistribution, UCalcRulesImplementation,
   UCalcAssocStructure, EZDSLBar, UCalcSpvStructScore, UCalcSpvCN2;

TYPE
    //****************
    //types de mesures
    //****************
    TEnumMeasureRuleInduction = (mriConfidence, mriMisclassification, mriAsymJMeasure);

    //**************************
    //paramtre pour les calculs
    //**************************
    TPrmFuncRuleInduction = class(TPrmFuncDistEval)
    //modalit de rfrence
    ClassRef: TTypeDiscrete;
    //poids pour la J-Measure asymtrique
    Weight: Double;
    end;


CONST
    //***********************
    //description des mesures
    //***********************
    STR_MEASURE_RULE_INDUCTION : array[TEnumMeasureRuleInduction] of string =
    ('Confidence Statistic','Misclassification Rate Statistic','Asymmetric J-Measure');

TYPE
   {*************************************}
   {* algorithme d'induction des rgles *}
   {*************************************}
   TCalcSpvRuleInduction = class(TCalcSpvCN2)
   private
   //Poids pour la J-Measure asymmtrique
   FWeight: double;
   //Paramtres de la fonction de calcul
   FPrmFunc: TPrmFuncRuleInduction;
   //ordre de passage des modalits
   FOrdValues: array of TTypeDiscrete;
   //les observations de la classe de rfrence
   FPositifs: TBooleanArray;
   //calculer l'ordre de passage des modalits
   procedure   calcOrdValues(dist: TTabFrequence);
   //calculer la rgles par dfaut
   procedure   calcDefaultRule(examples: TExamples);
   //indexer tous les positifs
   procedure   indexerPositifs(classe: TTypeDiscrete; positifs: TBooleanArray; examples: TExamples);
   protected
   //rcuprer les paramtres de calcul
   procedure   getParameters(); override;
   //chercher la rgle suivante -- renvoie NIL si pas de rgle possible
   function    searchNextRule(): TRuleProductionRule; override;
   //gestion des structures
   procedure   createStructures(); override;
   procedure   destroyStructures(); override;
   {apprentissage proprement dit}
   function    coreLearning(examples: TExamples): boolean; override;
   public
   procedure   getScore(example: integer; var postProba: TTabScore); override;
   function    getHTMLResults(): string; override;
   end;


implementation

USES
    UCompSpvRuleInduction, UStringAddBuffered, Math, Sysutils,
    UConstConfiguration, ULogFile, FMath;

{ TCalcSpvRuleInduction }

procedure TCalcSpvRuleInduction.calcDefaultRule(examples: TExamples);
var i, example: integer;
    rule: TRuleProductionRule;
    r: integer;
    trouve: boolean;
    conclusion: TTypeDiscrete;
begin
 //les individus qui ne sont couverts par aucune des rgles
 if assigned(FDefaultConclusion) then FDefaultConclusion.Free();
 FDefaultDistribution.ReInitialization();
 for i:= 1 to examples.Size do
  begin
   example:= examples.Number[i];
   trouve:= false;
   for r:= 0 to pred(FRules.count) do
    begin
     rule:= FRules.getRule(r);
     trouve:= rule.testExample(example);
     if (trouve = true) then break;
    end;
   //si pas couvert, l'intgrer dans la rgle par dfaut
   if (trouve = false)
    then FDefaultDistribution.IncrementCell(self.ClassAttribute.dValue[example]);
  end;
 //crer la rgle par dfaut
 conclusion:= FDefaultDistribution.getIndexMaxValue();
 FDefaultConclusion:= TRuleProductionRule.Create(FALSE,TRuleCondDiscrete.Create(self.ClassAttribute,ord(rcdIn),conclusion));
 FDefaultConclusion.setDistribution(FDefaultDistribution);
end;

procedure TCalcSpvRuleInduction.calcOrdValues(dist: TTabFrequence);
var i,j,jMin: integer;
    tmp: TTabFrequence;
    min: TFreqCountType;
    id: TTypeDiscrete;
begin
 tmp:= TTabFrequence.Create(dist.Size);
 tmp.Copy(dist);
 //initialiser avec 1, 2, 3, ...
 for i:= 1 to self.ClassAttribute.nbValues do FOrdValues[i]:= i;
 //trier selon les frquences - faisons simple, le nombre de modalits est faible toujours
 for i:= 1 to self.ClassAttribute.nbValues-1 do
  begin
   jMin:= i;
   min:= tmp.Value[jMin];
   for j:= i+1 to self.ClassAttribute.nbValues do
    begin
     if (tmp.Value[j] < min) then
      begin
       jMin:= j;
       min:= tmp.Value[jMin];
      end;
    end;
   //procder  l'change sur le vecteur des distributions
   tmp.Value[jMin]:= tmp.Value[i];
   tmp.Value[i]:= min;
   //et sur le vecteur des index
   id:= FOrdValues[i];
   FOrdValues[i]:= FOrdValues[jMin];
   FOrdValues[jMin]:= id;
  end;
 //don't forget
 tmp.Free();
end;

function TCalcSpvRuleInduction.coreLearning(examples: TExamples): boolean;
var k: integer;
    rule: TRuleProductionRule;
begin
 self.getParameters(); //rcuprer les paramtres de calcul
 self.buildAllItems(examples); //construire tous les items
 self.getRootExamples(examples); //construire la racine -- les exemples  traiter
 self.calcOrdValues(FRootDistribution); //ordonner les modalits par ordre de frquence croissante
 result:= true;
 TRY
 FPositifs:= TBooleanArray.Create(examples.Size);
 //pour chaque classe de focalisation, on ne traite pas la dernire qui devient la rgle par dfaut
 for k:= 1 to pred(self.ClassAttribute.nbValues) do
  begin
   //rcuperer la classe de focalisation
   FPrmFunc.ClassRef:= FOrdValues[k];
   //indexer tous les positifs
   self.indexerPositifs(FPrmFunc.ClassRef, FPositifs,examples);
   //raliser le processus d'apprentissage pour la classe FPrmFunc.ClassRef
   REPEAT
    rule:= self.SearchNextRule();
    if assigned(rule)
     then FRules.addRule(rule)
   UNTIL (rule = nil);
   //remettre tout le monde dans le pot pour la classe de focalisation suivante
   FCurExamples.SetAllTrue();
   FCurExamples.RefreshCount();
  end;
 FPositifs.Free();
 //puis calculer la distribution pour la rgle par dfaut
 self.calcDefaultRule(examples);
 EXCEPT
 result:= false;
 END;
end;

procedure TCalcSpvRuleInduction.createStructures;
begin
 inherited;
 //et le paramtrage de la fonction
 FPrmFunc:= TPrmFuncRuleInduction.Create();
 SetLength(FOrdValues, self.ClassAttribute.nbValues + 1);
end;

procedure TCalcSpvRuleInduction.destroyStructures;
begin
 inherited;
 FPrmFunc.Free();
 if (FOrdValues <> nil) then Finalize(FOrdValues);
end;

function TCalcSpvRuleInduction.getHTMLResults: string;
var s: string;
    i: integer;
    rule: TRuleProductionRule;
    buf: TBufString;
begin
 buf:= TBufString.Create();
 buf.BeginUpdate();
 buf.AddStr(format('<H3>Number of rules = %d</H3>',[FRules.Count]));
 buf.AddStr('<H3>Knowledge-based system</H3>');
 buf.AddStr(HTML_HEADER_TABLE_RESULT);
 buf.AddStr(HTML_TABLE_COLOR_HEADER_GRAY+'<TH>Antecedent</TH><TH>Consequent</TH><TH>Distribution</TH></TR>');
 //chaque rgle
 for i:= 0 to pred(FRules.count) do
  begin
   rule:= FRules.getRule(i);
   //
   s:= HTML_TABLE_COLOR_DATA_GRAY;
   //premisse
   s:= s+'<TD>';
   s:= s+'IF '+rule.Premisse.getHTMLDescription()+'</TD>';
   //conclusion
   s:= s+'<TD>'+rule.Conclusion.getHTMLDescription()+'</TD>';
   //distribution
   s:= s+'<TD align="right">'+rule.Distribution.getHTMLResultSimplified()+'</TD>';
   s:= s+'</TR>';
   //
   buf.AddStr(s);
  end;
 //rgle par dfaut
 if assigned(FDefaultConclusion)
  then buf.AddStr(HTML_TABLE_COLOR_DATA_GRAY+'<TD>(DEFAULT RULE)</TD><TD>'+
                  FDefaultConclusion.Conclusion.getHTMLDescription()+'</TD><TD align="right">'+FDefaultDistribution.getHTMLResultSimplified()+'</TD></TR>');
 buf.AddStr('</table>');
 buf.EndUpdate();
 result:= buf.BufS;
 buf.Free();
end;

procedure TCalcSpvRuleInduction.getParameters;
var prm: TOpPrmRuleInduction;
begin
 prm:= OpPrmSpv as TOpPrmRuleInduction;
 //type de mesure
 case prm.Measurement of
  mriConfidence: FMeasure:= VTConfidence;
  mriMisclassification: FMeasure:= VTMisclassification;
  else FMeasure:= AsymmetricJMeasure;
 end;
 //signif. level
 FSignifLevel:= prm.SignifLevel;
 //support
 FSupportMin:= prm.SupportMin;
 //Poids pour la J-Measure
 FWeight:= prm.Weight;
 //prparer le paramtrage de la fonction de calcul
 FPrmFunc.Weight:= FWeight;
end;

procedure TCalcSpvRuleInduction.getScore(example: integer;
  var postProba: TTabScore);
var rule: TRuleProductionRule;
    i: integer;
begin
 rule:= nil;
 postProba.raz();
 //passer en revue toutes les rgles !!!
 for i:= 0 to pred(FRules.Count()) do
  begin
   if FRules.getRule(i).testExample(example)
    then
     begin
      //si rgle dclenche, on ajoute la distribution...
      rule:= FRules.getRule(i);
      postProba.addFromTabFrequence(rule.Distribution);
     end;
  end;
 //si pas de rgle dclenche -- utilisation de la rgle par dfaut
 if not(assigned(rule))
  then postProba.recupFromTabFrequence(FDefaultConclusion.Distribution);
 //si vide, prendre la racine
 if (postProba[0] = 0) then postProba.recupFromTabFrequence(FRootDistribution);
end;
  

procedure TCalcSpvRuleInduction.indexerPositifs(classe: TTypeDiscrete; positifs: TBooleanArray;
  examples: TExamples);
var i: integer;
begin
 for i:= 1 to examples.Size do
  positifs.Flag[pred(i)]:= (self.ClassAttribute.dValue[examples.Number[i]] = classe);
 positifs.RefreshCount();
end;

function TCalcSpvRuleInduction.searchNextRule: TRuleProductionRule;
var covP,covC: TBooleanArray;
    rule: TRuleProductionRule;
    stopSearch: boolean;
    rootDist,curDist: TTabFrequence;
    item: TAssocItem;
    j,jMin: integer;
    minFit,Fit: double;
    pValue: double;
begin
 //individus de travail, copie de travail
 covP:= TBooleanArray.CreateFrom(FCurExamples);
 covP.RefreshCount();
 //distribution au dpart (racine) de la spcialisation
 rootDist:= TTabFrequence.CreateFromAtt(self.ClassAttribute,NIL);
 computeClassDistribution(covP,rootDist);
 //dist. courante durant la recherche
 covC:= TBooleanArray.Create(FCurExamples.Capacity);
 curDist:= TTabFrequence.CreateFromAtt(self.ClassAttribute,NIL);
 //rgle de travail
 rule:= TRuleProductionRule.create(TRUE,NIL);
 //boucler
 stopSearch:= false;
 minFit:= MATH.MaxDouble;
 while not(stopSearch) do
  begin
   jMin:= -1;
   //pour chaque couple "attribut-valeur" candidat
   for j:= 0 to pred(FDescriptorsItems.Count) do
    begin
     item:= FDescriptorsItems.Item[j];
     //vrifier si l'attribut est dj dans la rgle
     if rule.isDescriptorPresent(item.Attribute)
      //passer  l'attribut suivant
      then CONTINUE
      else
       begin
        //calculer les individus couverts
        covC.CopyFrom(covP);
        covC.AndArray(item.BoolArray);
        covC.RefreshCount();
        //et la distribution sur la classe
        computeClassDistribution(covC,curDist);
        //tester d'abord si a passe les rgles d'arrt (pre-pruning)
        if (curDist.Value[0] >= FSupportMin)
         then
          //est-ce qu'il est meilleur alors ?
          begin
           Fit:= FMeasure(rootDist,curDist,FPrmFunc);
           if (Fit < minFit)
            then
             begin
              minFit:= Fit;
              jMin:= j;
             end;
          end;
       end;
    end;
   //alors ?
   if (jMin>=0)
    then
     begin
      //ajouter l'item  la rgle
      item:= FDescriptorsItems.Item[jMin];
      rule.addCondition(TRuleCondDiscrete.Create(item.Attribute,ord(rcdIn),item.Value));
      //refresh des individus couverts
      covP.AndArray(item.BoolArray);
      covP.RefreshCount();//on a les individus couverts par la prmisse ici...
     end
    //sinon, il est temps de s'arrter
    else stopSearch:= TRUE;
  end;
 //mettre  jour les distributions
 if rule.isEmptyPremisse()
  then
   begin
    FreeAndNil(rule);
    curDist.Free();
   end
  else
   begin
    //vrifier que la rgle passe le seuil de significativit pour les critres statistiques
    pValue:= 0.0;
    if ((OpPrmSpv as TOpPrmRuleInduction).Measurement in [mriConfidence,mriMisclassification])
     then pValue:= FNorm(minFit);
    //calculer la distribution
    self.computeClassDistribution(covP,curDist);
    //a ne passe pas dans 2 cas : la classe majoritaire n'est pas celle que l'on veut OU la p-value est suprieure au seuil
    if (curDist.getUniqueIndexMaxValue() <> FPrmFunc.ClassRef) OR (pValue > FSignifLevel)
     then
      begin
       FreeAndNil(rule);
       curDist.Free();
      end
     else
      begin
       //calculer la conclusion de la rgle
       rule.setConclusion(TRuleCondDiscrete.Create(self.ClassAttribute,ord(rcdIn),getRuleConclusion(covP,curDist)));
       rule.setDistribution(curDist);//et ne pas dtruire curDist surtout !!!
       //on ne va liminer que les individus couverts et qui sont de la classe de focalisation
       //on les dtecte ici
       covP.AndArray(FPositifs);
       covP.RefreshCount();
       //TraceLog.WriteToLogFile(format('[RULE-INDUCTION] - removed examples = %d',[covP.Count]));
       //inverser la slection dans covP
       covP.ToggleAll();
       covP.RefreshCount();
       //les liminer de la slection courante
       FCurExamples.AndArray(covP);
       FCurExamples.RefreshCount();
       //TraceLog.WriteToLogFile(format('[RULE-INDUCTION] - remaining examples = %d',[FCurExamples.Count]));
       //******
      end;
   end;
 //renvoyer l rgle (qui est  nil si rien n'a t ajout
 result:= rule;
 //rz
 covP.Free();
 covC.Free();
 //new -- 29/10/2009 -- memory leak
 if (rootDist <> nil) then FreeAndNil(rootDist);
end;

end.
