(**********************************************************************)
(* UCalcStatDesPairedTest.pas - Copyright (c) 2005 Ricco RAKOTOMALALA *)
(**********************************************************************)

{
@abstract(Classes de calcul pour les tests sur K chantillons apparis)
@author(Ricco)
@created(04/08/2005)

Principale rfrence :
----------------------
Siegel & Castellan (Nonparametric Statistics for the Behavioral Science -- dition de 1988) -- chapitre 7 et 9

Deux mthodes seront implments :
(1) FRIEDMAN 2-way ANOVA by RANKS
(2) KENDALL COEFFICIENT OF CONCORDANCE

New -- 12/08/2008 -- Ajout de l'ANOVA pour Blocs complets alatoires (Randomized Complete Block Design)

}

unit UCalcStatDesKRelatedSamples;

interface

USES
    UDatasetDefinition,
    UDatasetExamples,
    UCalcStatDesRankForNonParametricStat;

TYPE
    //sa structure est trs diffrente des autres, il n'y a pas une variable en particulier  traiter
    //on recommence  zro donc
    TCalcSDKRelated = class(TObject)
    protected
    //liste des attributs
    FLstAtt : TLstAttributes;
    //tableau des rangs -- la dimension est k ou N selon la mthode
    FTabRanks:array of double;
    //correction pour les ex-aequos
    FTieCorrection: double;
    //la statistique
    FStat: double;
    //la p-value
    FPValue: double;
    //ddl
    FDDL: integer;
    //nombre d'observations
    FNbExamples: integer;
    //classe de calcul des rangs
    function    getClassComputingRanks(): TClassRankComputed; virtual; abstract;
    public
    //passer la liste de variables
    constructor create(prmLst: TLstAttributes); virtual;
    //destructeur
    destructor  destroy(); override;
    //calcule les stats, renvoie TRUE si OK
    function    computeStats(prmExamples: TExamples): boolean; virtual; abstract;
    //renvoyer les rsultats sous forme HTML
    function    getHTMLResults(): string; virtual; abstract;
    end;

    //test de Friedman -- p.174
    TCalcSDKRelatedFriedman = class(TCalcSDKRelated)
    protected
    //calcul effectif de la statistique de Friedman
    procedure   calcStatistics(prmExamples: TExamples); virtual;
    //classe de calcul des rangs
    function    getClassComputingRanks(): TClassRankComputed; override;
    public
    //calcule les stats, renvoie TRUE si OK
    function    computeStats(prmExamples: TExamples): boolean; override;
    //sortie
    function    getHTMLResults(): string; override;
    end;

    //W de concordance de Kendall -- p.262
    //exploiter la similitude avec le test de Friedman
    TCalcSDKRelatedWKendall = class(TCalcSDKRelatedFriedman)
    protected
    //la stat. du chi-2 associe
    FChi2: double;
    //moyenne du rho de Spearman
    FAvgRho: double;
    //calcul effectif de la statistique de Friedman
    procedure   calcStatistics(prmExamples: TExamples); override;
    //calcul des rangs
    function    getClassComputingRanks(): TClassRankComputed; override;  
    public
    //sortie
    function    getHTMLResults(): string; override;
    end;

    //Anova pour les blocs complets alatoires (Randomized Blocks Design)
    TCalcSDKRelatedAnova = class(TCalcSDKRelated)
    protected
    //http://eric.univ-lyon2.fr/~ricco/cours/cours/Comp_Pop_Tests_Parametriques.pdf
    //page 50, tableau 4.1
    FSce, FScb, FScrPrim, FSct, FCme, FCmb, FCmrPrim: double;
    //stat. pour les blocs
    FStatBlock, FPValueBlock: double;
    //calcul des rangs - inutile ici --> NIL
    function    getClassComputingRanks(): TClassRankComputed; override;
    public
    function    computeStats(prmExamples: TExamples): boolean; override;
    //sortie
    function    getHTMLResults(): string; override;
    end;

    //test Q de Cochran -- trs facile  construire
    //attention, les variables doivent toutes tre de type continu, mais binaires 0/1
    TCalcSDKRelatedQCochran = class(TCalcSDKRelated)
    protected
    //tableau des sommes, utiliss pour la statistique
    FSumRow, FSumCol: array of double;
    public
    //calcule les stats, renvoie TRUE si OK
    function    computeStats(prmExamples: TExamples): boolean; override;
    //sortie
    function    getHTMLResults(): string; override;
    //destructeur
    destructor  destroy(); override;    
    end;


implementation

uses
    Math,
    Sysutils,
    UDatasetImplementation,
    FMath, UConstConfiguration;

{ TCalcSDKRelated }

constructor TCalcSDKRelated.create(prmLst: TLstAttributes);
begin
 inherited Create();
 FLstAtt:= prmLst;
end;

destructor TCalcSDKRelated.destroy;
begin
 if (FTabRanks <> nil) then Finalize(FTabRanks);
 inherited destroy();
end;

{ TCalcSDKRelatedFriedman }

procedure TCalcSDKRelatedFriedman.calcStatistics(prmExamples: TExamples);
var cK,cN: double;
    s2Ranks,num,denom: double;
    k: integer;
begin
 //taille de tableau -- on se rfre  la terminologie de Siegel, p.179
 cN:= 1.0 * prmExamples.Size;
 ck:= 1.0 * FLstAtt.Count;
 //la somme au carr des rangs
 s2Ranks:= 0.0;
 for k:= 0 to pred(FLstAtt.Count) do
  s2Ranks:= s2Ranks + FTabRanks[k] * FTabRanks[k];
 //ddl.
 FDDL:= FLstAtt.Count - 1;
 //si tout va bien, on calcule la statistique
 //numrateur
 num:= 12.0 * s2Ranks - 3.0 * power(cN,2.0) * cK * power(cK+1.0,2.0);
 //dnominateur
 denom:= cN* cK * (cK + 1.0) + 1.0/(cK - 1.0) * (cN * cK - FTieCorrection);
 //la stat.
 if (denom > 0.0)
  then FStat:= num / denom
  else FStat:= 0.0;
 //p-value
 FPValue:= PKhi2(FDDL,FStat);
end;

function TCalcSDKRelatedFriedman.computeStats(
  prmExamples: TExamples): boolean;
var attRanks: TAttContinue;
    lstSorted: TExamples;
    ranks: TRankComputed;
    i,k: integer;
    correction: double;
begin
 FNbExamples:= prmExamples.Size;
 //tableau -- somme des rangs (s'assurer qu'il sera vide et rempli de zros !)
 if (FTabRanks <> nil) then Finalize(FTabRanks);
 setLength(FTabRanks,FLstAtt.Count);
 //variables de calculs -- un peu tir par les cheveux mais a a le mrite de profiter  fond des structures existantes
 //!\on prend le tableau de donnes de travers ( l'horizontale)
 lstSorted:= TExamples.Create(FLstAtt.Count);
 attRanks:= TAttContinue.Create('_att_',FLstAtt.Count);
 ranks:= getClassComputingRanks.create(attRanks);
 TRY
  result:= TRUE;
  TRY
   FTieCorrection:= 0.0;
   //pour chaque observation
   for i:= 1 to prmExamples.Size do
    begin
     //remplir la variable temporaire
     for k:= 0 to pred(FLstAtt.Count) do
      attRanks.cValue[succ(k)]:= FLstAtt.Attribute[k].cValue[prmExamples.Number[i]];
     //recalculer le tableau de rangs
     lstSorted.fillValues();
     lstSorted.QuickSortBy(attRanks);
     //mme pour les ex-aequos de taille 1, on corrige (!) -- Siegel, p.179 -- dernier paragraphe
     //calcul en 2 temps pour dbogage et affichage dans le log.
     correction:= ranks.computeRank(lstSorted);
     //
     FTieCorrection:= FTieCorrection + correction;
     //faire la somme des rangs
     for k:= 0 to pred(FLstAtt.Count) do
       FTabRanks[k]:= FTabRanks[k] + ranks.getRank(succ(k));
    end;
   //calculer les statistiques --> Stat, DDL, et p-value
   self.calcStatistics(prmExamples);
  EXCEPT
   result:= FALSE;
  END;
 FINALLY
 ranks.Free();
 lstSorted.Free();
 attRanks.Free();
 END;
end;

function TCalcSDKRelatedFriedman.getClassComputingRanks: TClassRankComputed;
begin
 result:= TRankComputedKRelatedFriedman; 
end;

function TCalcSDKRelatedFriedman.getHTMLResults: string;
var s: string;
    k: integer;
begin
 s:= '<H3>Results</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>RANKS</TH><TH>Friedman Statistic</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY;
 //description des rangs
 s:= s + '<TD valign="top">';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_BLUE + '<TH>Att.</TH><TH>Sum(Ranks)</TH><TH>Mean(Ranks)</TH></TR>';
 for k:= 0 to pred(FLstAtt.Count) do
  s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TD>%s</TD><TD align=right>%.1f</TD><TD align="right">%.4f</TD></TR>',
                                        [FLstAtt.Attribute[k].Name,FTabRanks[k],FTabRanks[k]/(1.0*FNbExamples)]);
 s:= s + '</table>';
 s:= s + '</TD>';
 //descritpion de la statistique
 s:= s + '<TD valign="top">';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_BLUE + '<TH>Stat.</TH><TH>Value</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD>Frideman Fr</TD><TD align="right">%.5f</TD></TR>',[FStat]);
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD>d.f.</TD><TD align="right">%d</TD></TR>',[FDDL]);
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD>p-value</TD><TD align="right" %s>%.5f</TD></TR>',[colorCodeForPValue(FPValue),FPValue]);
 s:= s + '</table>';
 s:= s+'</TD>';
 //finaliser le tableau
 s:= s + '</TR>';
 s:= s + '</table>';
 //
 result:= s;
end;

{ TCalcSDKRelatedWKendall }

procedure TCalcSDKRelatedWKendall.calcStatistics(prmExamples: TExamples);
var cK,cN: double;
    s2Ranks,num,denom: double;
    k: integer;
begin
 //terminologie Siegel, p.262 -- on prend le tableau dans l'autre sens !
 cK:= 1.0 * prmExamples.Size;  //nombre de ligne = k
 cN:= 1.0 * FLstAtt.Count; //nombre de colonnes = N
 //la somme au carr des rangs
 s2Ranks:= 0.0;
 for k:= 0 to pred(FLstAtt.Count) do
  s2Ranks:= s2Ranks + FTabRanks[k] * FTabRanks[k];
 //la statistique, p.266 (9.18b)
 num:= 12.0 * s2Ranks - 3.0 * power(cK,2.0) * cN * power(cN + 1.0,2.0);
 denom:= power(cK,2.0) * cN * (power(cN,2.0) - 1.0) - cK * FTieCorrection;
 if (denom > 0.0)
  then FStat:= num / denom
  else FStat:= 0.0;
 //moyenne des rho de Spearman (calculs 2  2) -> Siegel, p. 262 , eq. 9.16
 FAvgRho:= (cK * FStat - 1.0) / (cK - 1.0);
 //la stat transforme -> chi-2 pour le test (eq. 9.19)
 FChi2:= cK * (cN - 1.0) * FStat;
 //ddl -> (N-1)
 FDDL:= FLstAtt.Count - 1;
 //p-value
 FPValue:= PKhi2(FDDL,FChi2);
end;

function TCalcSDKRelatedWKendall.getClassComputingRanks: TClassRankComputed;
begin
 result:= TRankComputedKRelatedWKendall;
end;

function TCalcSDKRelatedWKendall.getHTMLResults: string;
var s: string;
    k: integer;
begin
 s:= '<H3>Results</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>RANKS</TH><TH>Concordance Statistic</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY;
 //description des rangs
 s:= s + '<TD valign="top">';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_BLUE + '<TH>Att.</TH><TH>Sum(Ranks)</TH><TH>Mean(Ranks)</TH></TR>';
 for k:= 0 to pred(FLstAtt.Count) do
  s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TD>%s</TD><TD align=right>%.1f</TD><TD align="right">%.4f</TD></TR>',
                                        [FLstAtt.Attribute[k].Name,FTabRanks[k],FTabRanks[k]/(1.0*FNbExamples)]);
 s:= s + '</table>';
 s:= s + '</TD>';
 //descritpion de la statistique
 s:= s + '<TD valign="top">';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_BLUE + '<TH>Stat.</TH><TH>Value</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD>Kendall W</TD><TD align="right">%.5f</TD></TR>',[FStat]);
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD>Average(Spearman''s Rho)</TD><TD align="right">%.5f</TD></TR>',[FAvgRho]);
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD>Chi-2</TD><TD align="right">%.5f</TD></TR>',[FChi2]);
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD>d.f.</TD><TD align="right">%d</TD></TR>',[FDDL]);
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD>p-value</TD><TD align="right" %s>%.5f</TD></TR>',[colorCodeForPValue(FPValue),FPValue]);
 s:= s + '</table>';
 s:= s+'</TD>';
 //finaliser le tableau
 s:= s + '</TR>';
 s:= s + '</table>';
 //
 result:= s;
end;

{ TCalcSDKRelatedAnova }

function TCalcSDKRelatedAnova.computeStats(
  prmExamples: TExamples): boolean;
var i,k: integer;
    gN,gK: integer;
    vec: array of double;
    xbarre: double;
begin
 result:= true;
 FStat:= 0.0;
 FPValue:= 1.0;
 try
  if ((self.FLstAtt.Count-1)*(prmExamples.Size-1) > 0)
   then
    begin
     FNbExamples:= prmExamples.Size;
     gN:= FNbExamples;
     gK:= self.FLstAtt.Count;
     //calculer la moyenne globale
     xbarre:= 0.0;
     for i:= 1 to gN do for k:= 0 to pred(gK) do xbarre:= xbarre + self.FLstAtt.Attribute[k].cValue[prmExamples.Number[i]];
     xbarre:= xbarre/(gN*gK);
     //calculer la sct
     FSct:= 0.0;
     for i:= 1 to gN do for k:= 0 to pred(gK) do FSct:= FSct + SQR(self.FLstAtt.Attribute[k].cValue[prmExamples.Number[i]] - xbarre);
     //vecteur de moyenne pour la sce
     setlength(vec,gK);
     for k:= 0 to pred(gK) do
      begin
       vec[k]:= 0.0;
       for i:= 1 to gN do vec[k]:= vec[k] + self.FLstAtt.Attribute[k].cValue[prmExamples.Number[i]];
       vec[k]:= vec[k]/(1.0 * gN);
      end;
     FSce:= 0.0;
     for k:= 0 to pred(gK) do FSce:= FSce + sqr(vec[k] - xbarre);
     FSce:= (1.0 * gN) * FSce;
     //blocs
     setlength(vec,gN+1);
     for i:= 1 to gN do
      begin
       vec[i]:= 0.0;
       for k:= 0 to pred(gK) do vec[i]:= vec[i] + self.FLstAtt.Attribute[k].cValue[prmExamples.Number[i]];
       vec[i]:= vec[i] / (1.0 * gK);
      end;
     FScb:= 0.0;
     for i:= 1 to gN do FScb:= FScb + sqr(vec[i] - xbarre);
     FScb:= (1.0 * gK) * FScb;
     //par diffrence
     FScrPrim:= (FSct - (FSce + FScb));
     //la statistique
     if (FScrPrim > 0.0)
      then
       begin
        FCme:= FSce/(-1.0 + gK);
        FCmb:= FScb/(-1.0 + gN);
        FCmrPrim:= FScrPrim/((-1.0+gK)*(-1.0+gN));
        FStat:= FCme/FCmrPrim;
        FPValue:= PSnedecor(gK-1,(gN-1)*(gK-1),FStat);
        //pour les blocs aussi
        FStatBlock:= FCmb/FCmrPrim;
        FPValueBlock:= PSnedecor(gN-1,(gN-1)*(gK-1),FStatBlock);
       end;
     //vider
     Finalize(vec);
    end;
 except
 result:= false;
 end;
end;

function TCalcSDKRelatedAnova.getClassComputingRanks: TClassRankComputed;
begin
 result:= nil;
end;

function TCalcSDKRelatedAnova.getHTMLResults: string;
var s: string;
begin
 s:= '<H3>Anova for Randomized Complete Block Design</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH width=60>Source</TH><TH width=60>Sum.Square</TH><TH width=60>d.f.</TH><TH width=60>Mean.Square</TH><TH width=60>F value</TH><TH width=60>Pr > F</TH></TR>';
 //traitement
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD align=left %s>Treatment</TD>',[HTML_BGCOLOR_HEADER_GRAY]);
 s:= s + format('<TD align=center>%.4f</TD><TD align=center>%d</TD><TD align=center>%.4f</TD><TD align=center>%.4f</TD><TD align=center>%.5f</TD>',
                [FSce,self.FLstAtt.Count-1,FCme,FStat,FPValue]);
 s:= s + '</TR>';
 //blocs
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD align=left %s>Block</TD>',[HTML_BGCOLOR_HEADER_GRAY]);
 s:= s + format('<TD align=center>%.4f</TD><TD align=center>%d</TD><TD align=center>%.4f</TD><TD align=center>%.4f</TD><TD align=center>%.5f</TD>',
                [FScb,FNbExamples-1,FCmb,FStatBlock,FPValueBlock]);
 s:= s + '</TR>';
 //error
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD align=left %s>Error</TD>',[HTML_BGCOLOR_HEADER_GRAY]);
 s:= s + format('<TD align=center>%.4f</TD><TD align=center>%d</TD><TD align=center>%.4f</TD><TD align=center>-</TD><TD align=center>-</TD>',
                [FScrPrim,(FNbExamples-1)*(self.FLstAtt.Count-1),FCmrPrim]);
 s:= s + '</TR>';
 //total
 s:= s + HTML_TABLE_COLOR_DATA_GRAY + format('<TD align=left %s>Total</TD>',[HTML_BGCOLOR_HEADER_GRAY]);
 s:= s + format('<TD align=center>%.4f</TD><TD align=center>%d</TD><TD align=center>-</TD><TD align=center>-</TD><TD align=center>-</TD>',
                [FSct,FNbExamples*self.FLstAtt.Count-1]);
 s:= s + '</TR>';
 //and then...
 s:= s + '</table>';
 result:= s;
end;

{ TCalcSDKRelatedQCochran }

function TCalcSDKRelatedQCochran.computeStats(
  prmExamples: TExamples): boolean;
var i,k: integer;
    num,denom,gK,S,r: double;
begin
 result:= true;
 TRY
 setLength(FSumRow,prmExamples.Size);
 setLength(FSumCol,FLstAtt.Count);
 //un peu bourrin mais au moins c'est scure...
 //sommes en ligne
 for i:= 1 to prmExamples.Size do
  begin
   FSumRow[pred(i)]:= 0.0;
   for k:= 0 to pred(FLstAtt.Count) do
    FSumRow[pred(i)]:= FSumRow[pred(i)] + FLstAtt.Attribute[k].cValue[prmExamples.Number[i]];
  end;
 //sommes en colonne
 S:= 0.0;
 for k:= 0 to pred(FLstAtt.Count) do
  begin
   FSumCol[k]:= 0.0;
   for i:= 1 to prmExamples.Size do
    FSumCol[k]:= FSumCol[k] + FLstAtt.Attribute[k].cValue[prmExamples.Number[i]];
   S:= S + FSumCol[k];
  end;
 //construire la statistique
 gK:= 1.0 * FLstAtt.Count;
 r:= S/gK;
 //numrateur
 num:= 0.0;
 for k:= 0 to pred(FLstAtt.Count) do
  num:= num + power(FSumCol[k]-r,2.0);
 //dnominateur
 denom:= 0.0;
 for i:= 0 to pred(prmExamples.Size) do
  denom:= denom + FSumRow[i] * (gK - FSumRow[i]);
 //ddl
 FDdl:= FLstAtt.Count-1;
 //stat.
 FStat:= 0.0;
 if (denom > 0.0) then FStat:= gK * (gK - 1.0) * num / denom;
 //p-value
 FPValue:= PKhi2(FDDL,FStat);
 EXCEPT
 result:= false;
 END;
end;

destructor TCalcSDKRelatedQCochran.destroy;
begin
 setLength(FSumRow,0);
 setLength(FSumCol,0);
 inherited;
end;

function TCalcSDKRelatedQCochran.getHTMLResults: string;
var s: string;
    k: integer;
begin
 s:= '<H3>Results</H3>';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_GRAY + '<TH>Variables</TH><TH>Cochran''s Q Statistic</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_GRAY;
 //description des rangs
 s:= s + '<TD valign="top">';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_BLUE + '<TH>Att.</TH><TH>Sum(Values)</TH></TR>';
 for k:= 0 to pred(FLstAtt.Count) do
  s:= s + HTML_TABLE_COLOR_HEADER_GRAY + format('<TD>%s</TD><TD align=right>%.0f</TD></TR>',
                                        [FLstAtt.Attribute[k].Name,FSumCol[k]]);
 s:= s + '</table>';
 s:= s + '</TD>';
 //descritpion de la statistique
 s:= s + '<TD valign="top">';
 s:= s + HTML_HEADER_TABLE_RESULT;
 s:= s + HTML_TABLE_COLOR_HEADER_BLUE + '<TH>Stat.</TH><TH>Value</TH></TR>';
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD>Conchran''s Q</TD><TD align="right">%.5f</TD></TR>',[FStat]);
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD>d.f.</TD><TD align="right">%d</TD></TR>',[FDDL]);
 s:= s + HTML_TABLE_COLOR_DATA_BLUE + format('<TD>p-value</TD><TD align="right" %s>%.5f</TD></TR>',[colorCodeForPValue(FPValue),FPValue]);
 s:= s + '</table>';
 s:= s+'</TD>';
 //finaliser le tableau
 s:= s + '</TR>';
 s:= s + '</table>';
 //
 result:= s;
end;

end.
